/*****
*
* Copyright (C) 2001 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/


#include <stdio.h>

#include "detect.h"


typedef struct macaddr {
        uint32_t ipaddr;
        uint8_t macaddr[6];
        
        struct macaddr *next;
} macaddr_t;


static macaddr_t *macaddr_list[1024] = { NULL };


static int check_directed = 0;
static plugin_detect_t plugin;
static subscribtion_t subscribtion[] = {
        { p_arp, NULL },
	{ p_end, NULL },
};



__inline__ static unsigned int hash_addr(uint32_t ipaddr) 
{
        return ipaddr & (1024 - 1);
}



static int parse_mac_address(const char *string, uint8_t *addr) 
{
        int ret;
        uint32_t tmp[6];
        
        ret = sscanf(string, "%02x:%02x:%02x:%02x:%02x:%02x",
                     &tmp[0], &tmp[1], &tmp[2], &tmp[3], &tmp[4], &tmp[5]);
        
        if ( ret < 6 ) {
                log(LOG_ERR, "Argument is not a valid MAC address.\n");
                return -1;
        }
        
        addr[0] = (uint8_t) tmp[0];
        addr[1] = (uint8_t) tmp[1];
        addr[2] = (uint8_t) tmp[2];
        addr[3] = (uint8_t) tmp[3];
        addr[4] = (uint8_t) tmp[4];
        addr[5] = (uint8_t) tmp[5];
        
        return 0;
}



static int add_hash_entry(const char *ipaddr, const char *macaddr) 
{
        int ret;
        macaddr_t *addr;
        unsigned int key;
        
        addr = malloc(sizeof(macaddr_t));
        if ( ! addr ) {
                log(LOG_ERR, "memory exhausted.\n");
                return -1;
        }

        addr->next = NULL;
        
        addr->ipaddr = inet_addr(ipaddr);
        if ( addr->ipaddr < 0 ) {
                log(LOG_ERR, "Argument is not a valid IP address.\n");
                goto err;
        }
        
        ret = parse_mac_address(macaddr, addr->macaddr);
        if ( ret < 0 ) 
                goto err;

        key = hash_addr(addr->ipaddr);

        addr->next = macaddr_list[key];
        macaddr_list[key] = addr;
        
        return 0;

 err:
        free(addr);
        return -1;
}



static void check_arp_request(packet_container_t *pc, etherhdr_t *ether, etherarphdr_t *arp) 
{
        int ret;
        static uint8_t bcast[6] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

        if ( check_directed && memcmp(ether->ether_dhost, bcast, sizeof(ether->ether_shost)) ) 
                plugin_do_report((plugin_generic_t *)&plugin, pc,
                                 maybe_faked, "Directed ARP request",
                                 "ARP request directly directed at a host (not broadcasted).\n");
        
        ret = memcmp(ether->ether_shost, arp->arp_sha, sizeof(ether->ether_shost));
        if ( ret != 0 )
                plugin_do_report((plugin_generic_t *)&plugin, pc, maybe_faked, "ARP address mismatch",
                                 "Ethernet source address doesn't match ARP sender address.\n");
}




static void check_arp_reply(packet_container_t *pc, etherhdr_t *ether, etherarphdr_t *arp)
{
        int ret;
        
        ret = memcmp(ether->ether_shost, arp->arp_sha, sizeof(ether->ether_shost));
        if ( ret != 0 )
                plugin_do_report((plugin_generic_t *)&plugin, pc, maybe_faked, "ARP address mismatch",
                                 "Ethernet source address doesn't match ARP sender address.\n");

        ret = memcmp(ether->ether_dhost, arp->arp_tha, sizeof(ether->ether_dhost));
        if ( ret != 0 )
                plugin_do_report((plugin_generic_t *)&plugin, pc, maybe_faked, "ARP address mismatch",
                                 "Ethernet destination address doesn't match ARP target adress.\n");
}



static void check_cache_overwrite(packet_container_t *pc, etherhdr_t *ether, etherarphdr_t *arp) 
{
        uint32_t spa;
        macaddr_t *tmp;
        unsigned int key;

        spa = (*(uint32_t *) arp->arp_spa);
        key = hash_addr(spa);

        for ( tmp = macaddr_list[key]; tmp != NULL; tmp = tmp->next ) {                
                if ( tmp->ipaddr != spa )
                        continue;
                
                if ( memcmp(ether->ether_shost, tmp->macaddr, 6) != 0 )
                        plugin_do_report((plugin_generic_t *)&plugin, pc, maybe_faked,
                                         "Arp cache overwrite",
                                         "Attempted ARP cache overwrite attack : Source Ethernet address \n"
                                         "is not the same as real hardware address.\n");
                
                if ( memcmp(arp->arp_sha, tmp->macaddr, 6) != 0 ) 
                        plugin_do_report((plugin_generic_t *)&plugin, pc, maybe_faked,
                                         "ARP cache overwrite",
                                         "Attempted ARP cache overwrite attack : Source Hardware Address \n"
                                         "is not the same as real hardware address.\n");

                break;
        }
}




static void arpcheck_run(packet_container_t *packet) 
{
        etherhdr_t *ether;
        etherarphdr_t *arp = packet->packet[packet->depth].p.arp_hdr;

        if ( packet->packet[packet->depth - 1].proto != p_ether )
            return;

        if ( ntohs(arp->ea_hdr.ar_hrd) != ARPHRD_ETHER || 
             ntohs(arp->ea_hdr.ar_pro) != ETHERTYPE_IP )
                return;

        ether = packet->packet[packet->depth - 1].p.ether_hdr;
        
        switch ( ntohs(packet->packet[packet->depth].p.arp_hdr->ea_hdr.ar_op) ) {

        case ARPOP_REQUEST:
                check_arp_request(packet, ether, arp);
                break;

        case ARPOP_REPLY:
                check_arp_reply(packet, ether, arp);
                check_cache_overwrite(packet, ether, arp);
                break;
        }
}



static void print_help(const char *optarg) 
{
        fprintf(stderr, "Usage for %s :\n", plugin.name);
        fprintf(stderr, "\t -w --arpwatch <ipaddr> <macaddr> Check this address for cache overwrite attack.\n");
        fprintf(stderr, "\t -d --directed (default=%d) Check for directed ARP request.\n\n",
                check_directed);
}



static void set_directed(const char *optarg) 
{
        check_directed = 1;
}



static void set_watch(const char *optarg) 
{
        int ret;
        char *ipaddr, *macaddr;
        
        if ( ! optarg ) {
                log(LOG_ERR, "Argument should be: <ip> <macaddr>.\n");
                return;
        }

        ipaddr = strdup(optarg);
        if ( ! ipaddr ) {
                log(LOG_ERR, "memory exhausted.\n");
                return;
        }
        
        macaddr = strchr(ipaddr, ' ');
        if ( ! macaddr ) {
                log(LOG_ERR, "Ip address should be followed by a MAC address.\n");
                free(ipaddr);
                return;
        }

        *macaddr = 0;
        macaddr++;
        
        ret = add_hash_entry(ipaddr, macaddr);
        if ( ret < 0 )
                log(LOG_ERR, "error adding entry to the hash table.\n");
}




int plugin_init(unsigned int id)
{
        plugin_option_t opts[] = {
                { "arpwatch", required_argument, NULL, 'w', set_watch  },
                { "directed", no_argument, NULL, 'd', set_directed     },
                { "help", no_argument, NULL, 'h', print_help           },
                { 0, 0, 0, 0 },
        };
        
        plugin_set_name(&plugin, "ArpSpoof");        
        plugin_set_name(&plugin, "ArpSpoof");
        plugin_set_author(&plugin, "Yoann Vandoorselaere");
        plugin_set_contact(&plugin, "yoann@mandrakesoft.com");
        plugin_set_desc(&plugin, "Check for several ARP attack.");
        plugin_set_running_func(&plugin, arpcheck_run);
        plugin_set_subscribtion(&plugin, subscribtion);

        plugin_config_get((plugin_generic_t *)&plugin, opts, PRELUDE_CONF);
        
        return plugin_register((plugin_generic_t *) &plugin);
}
