/*****
*
* Copyright (C) 1998, 1999, 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <dlfcn.h>
#include <errno.h>
#include <limits.h> /* for NAME_MAX */
#include <sys/time.h>

#include "packet.h"

#include "list.h"
#include "plugin-common.h"
#include "plugin-common-prv.h"
#include "plugin-detect.h"
#include "plugin-protocol.h"
#include "common.h"


LIST_HEAD(tcp_plugins_list);
LIST_HEAD(udp_plugins_list);
LIST_HEAD(ip_plugins_list);
LIST_HEAD(ipencap_plugins_list);
LIST_HEAD(data_plugins_list);
LIST_HEAD(icmp_plugins_list);
LIST_HEAD(igmp_plugins_list);
LIST_HEAD(arp_plugins_list);
LIST_HEAD(rarp_plugins_list);
LIST_HEAD(ether_plugins_list);
LIST_HEAD(all_plugins_list);



static void register_to_plugin_provided_protocol(plugin_container_t *pc, const char *proto) 
{
        plugin_protocol_t *p;

        p = protocol_plugins_search(proto);
        if ( ! p )
                return;

        log(LOG_INFO, "%s", proto);
        
        plugin_register_for_use(pc, &p->detect_plugin_list, proto);
}




/*
 * Try to register plugin 'plugin' to protocol 'wanted_proto',
 * return -1 on error, 0 on success.
 */
static void register_to_internal_protocol(plugin_container_t *pc, int wanted_proto)
{
        int i;
        static struct {
                char *proto;
                proto_enum_t type;
                struct list_head *plugin_list;
        } tbl[] = {
                { "IP", p_ip, &ip_plugins_list                },
                { "IPENCAP", p_ipencap, &ipencap_plugins_list },
                { "TCP", p_tcp, &tcp_plugins_list             },
                { "UDP", p_udp, &udp_plugins_list             },
                { "ICMP", p_icmp, &icmp_plugins_list          },
                { "IGMP", p_igmp, &igmp_plugins_list          },
                { "ARP", p_arp, &arp_plugins_list             },
                { "RARP", p_rarp, &rarp_plugins_list          },
                { "ETHER", p_ether, &ether_plugins_list       },
                { "DATA", p_data, &data_plugins_list          },
                { "ALL", p_all, &all_plugins_list             },
                { NULL, 0, NULL                               }
        };
        
        for ( i = 0; tbl[i].proto != NULL; i++ )
                if ( tbl[i].type == wanted_proto ) {
                        log(LOG_INFO, "%s", tbl[i].proto);
                        plugin_register_for_use(pc, tbl[i].plugin_list, tbl[i].proto);
                        return;
                }

        log(LOG_ERR,"plugin %s has unknow type %d.\n", pc->plugin->name, wanted_proto);
}



/*
 * Subscribe plugins to it's specified protocol,
 * if the plugin need to subscribe to > 1 protocol :
 * we need to copy it.
 */
static int plugin_subscribe(plugin_container_t *pc)
{
	int i;
        plugin_detect_t *plugin = (plugin_detect_t *) pc->plugin;
        
        log(LOG_INFO, "\t\t%s subscribing to : \"[", plugin->name);
        
	for (i = 0; plugin->subscribtion[i].type != p_end; i++) {
                if ( i > 0 ) 
                        log(LOG_INFO, ",");

                if ( plugin->subscribtion[i].type != p_external ) 
                        register_to_internal_protocol(pc, plugin->subscribtion[i].type);
                else
                        register_to_plugin_provided_protocol(pc, plugin->subscribtion[i].filter);
        }

        log(LOG_INFO, "]\".\n");

        return 0;
}


/*
 * Start all plugins of kind 'list'.
 */
extern int depth;

void detect_plugins_run(packet_container_t *packet, struct list_head *list, char *proto)
{
        struct list_head *tmp;
        plugin_container_t *pc;

        list_for_each(tmp, list) {
                pc = list_entry(tmp, plugin_container_t, ext_list);
                plugin_run(pc, plugin_detect_t, packet);
        }
}



/*
 * Open the plugin directory (dirname),
 * and try to load all plugins located int it.
 */
int detect_plugins_init(const char *dirname)
{
        int ret;

        ret = plugin_load_from_dir(dirname, plugin_subscribe);
        if ( ret < 0 ) {
                log(LOG_ERR, "couldn't load plugin subsystem.\n");
                return -1;
        }

        return 0;
}


