/*****
*
* Copyright (C) 1998, 1999, 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>

#include "pcap.h"
#include "common.h"
#include "list.h"
#include "packet.h"
#include "ip_fragment.h"
#include "packet-decode.h"
#include "optparse.h"
#include "recycler.h"
#include "alert.h"
#include "alert-prv.h"
#include "plugin-common.h"
#include "plugin-detect.h"
#include "plugin-protocol.h"
#include "timer.h"
#include "rqueue.h"

#include "rules.h"
#include "rules_default.h"

extern struct list_head tcp_plugins_list;
extern struct list_head udp_plugins_list;
extern struct list_head ip_plugins_list;
extern struct list_head ipencap_plugins_list;
extern struct list_head data_plugins_list;
extern struct list_head icmp_plugins_list;
extern struct list_head igmp_plugins_list;
extern struct list_head arp_plugins_list;
extern struct list_head rarp_plugins_list;
extern struct list_head ether_plugins_list;
extern struct list_head all_plugins_list;

static unsigned int caplen;
static const struct pcap_pkthdr *pcap_hdr;
static const unsigned char *snapend;
packet_container_t *packet;

/*
 * recyclers.
 */
static recycler_t *packet_recycler;


inline static packet_container_t *packet_new(recycler_t *r, const unsigned char *p) 
{
        packet_container_t *pkt;

        pkt = malloc(sizeof(*pkt));
        if ( ! pkt )
                return NULL;

        pkt->protocol_plugin_data = NULL;
        pkt->protocol_plugin_id = -1;
        pkt->application_layer_depth = -1;
        pkt->transport_layer_depth = -1;
        pkt->network_layer_depth = -1;
        pkt->depth = -1;
        pkt->refcount = 0;
        pkt->allocated_data = NULL;
        pkt->captured_data = p;

        packet_lock(pkt);
        
        return pkt;
}



static void SliceAndStoreDataPkt(u_char *p)
{
        int len;
        
        /*
         * Don't issue warning if there is no data,
         * it can be absolutly normal. This rule doesn't apply if the
         * allocated_data pointer is set, because we can't check this
         * pointer against snapend.
         */
        if ( ! packet->allocated_data && p + 1 > snapend ) 
                return;

        /*
         * Run the protocol plugins. If the return value
         * is positive, one of them matched the payload of this packet.
         * The returned value is the len it handled before the real payload.
         */
        len = protocol_plugins_run(packet, p, caplen);
        if ( len < 0 )
                len = 0;
        
        /*
         * start data detection plugin *only* on the remaining data.
         */
        packet_add_header(packet, unsigned char, p, caplen - len, p_data, data);
        packet->application_layer_depth = packet->depth;
        
        if ( caplen - len > 0 ) 
                detect_plugins_run(packet, &data_plugins_list, "data");
        

        /*
         * Then, we want the whole packet data (including protocol plugin
         * data) dumped at reporting time. correct the header.
         */
        packet->packet[packet->depth].len = caplen;
}


static void SliceAndStoreTcpOptions(u_char *p, int len) 
{
        len = tcp_optdump(packet, p, len);
        if ( len > 0 ) 
                packet_add_header(packet, unsigned char, p, len, p_tcpopts, opts);
}



static void SliceAndStoreIpOptions(u_char *p, int len)
{
        len = ip_optdump(packet, p, len);
        if ( len > 0 )
                packet_add_header(packet, unsigned char, p, len, p_ipopts, opts);
}



static void SliceAndStoreTcpPkt(u_char *p)
{
        int hlen, len;
        tcphdr_t *tcp;
        
        if ( caplen < sizeof(tcphdr_t) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated TCP header",
                                  "truncated TCP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(tcphdr_t), caplen);
                return;
        }

        packet_add_header(packet, tcphdr_t, p, sizeof(tcphdr_t), p_tcp, tcp);
        packet->transport_layer_depth = packet->depth;
        tcp = packet->packet[packet->depth].p.tcp;
        
        hlen = tcp->th_off * 4;
        if ( hlen > caplen ) {
                prelude_do_report(packet, maybe_faked,
                                  "bad TCP header length",
                                  "bad TCP header length (hlen=%d, caplen=%d).\n\n ",
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", hlen, caplen);
                return;
        }

        caplen -= hlen;
        
        detect_plugins_run(packet, &tcp_plugins_list, "tcp");
        
        len = hlen - sizeof(tcphdr_t);
        if ( len > 0 )
                SliceAndStoreTcpOptions(p + sizeof(tcphdr_t), len);
                
        SliceAndStoreDataPkt(p + hlen);

#if 0
        tcp_stream_store(packet, packet->packet[depth_ip].p.ip,
                         packet->packet[depth_tcp].p.tcp, caplen);
#endif
        signature_engine_process_packet(signature_engine_get_tcp_root(), packet);
}

static void SliceAndStoreUdpPkt(u_char *p)
{
        /*
         * Sanity check.
         */
        if ( caplen < sizeof(udphdr_t) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated UDP",
                                  "truncated UDP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(udphdr_t), caplen);
                 return;
        }

        packet_add_header(packet, udphdr_t, p, sizeof(udphdr_t), p_udp, udp_hdr);
        caplen -= sizeof(udphdr_t);
        packet->transport_layer_depth = packet->depth;
        
        detect_plugins_run(packet, &udp_plugins_list, "udp");

        SliceAndStoreDataPkt(p + sizeof(udphdr_t));

	signature_engine_process_packet(signature_engine_get_udp_root(), packet);
}

static void SliceAndStoreIcmpPkt(unsigned char *p)
{
        int len;
        uint8_t type;
        uint8_t type_tbl[] = { 8,  /* type 0 */
                               0,  /* type 1 */
                               0,  /* type 2 */
                               36, /* type 3 */
                               36, /* type 4 */
                               36, /* type 5 */
                               0,  /* type 6 */
                               0,  /* type 7 */
                               8,  /* type 8 */
                               0,  /* type 9 router advertissement message
                                    * require extenssible struct */
                               
                               8,  /* type 10 */
                               36, /* type 11 */
                               36, /* type 12 */
                               20, /* type 13 */
                               20, /* type 14 */
                               8,  /* type 15 */
                               8,  /* type 16 */
                               12, /* type 17 address mask request */
                               12, /* type 18 address mask reply   */
        };

        
        if ( caplen < ICMP_MINLEN ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated ICMP",
                                  "truncated ICMP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", ICMP_MINLEN, caplen);
                return;
        }

        type = (((const icmphdr_t *)p)->icmp_type);
        if ( type < sizeof(type_tbl) ) {
                len = (type_tbl[type] != 0) ? type_tbl[type] : ICMP_MINLEN;
        } else
                len = ICMP_MINLEN;

        if ( caplen < len ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated ICMP",
                                  "truncated ICMP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", len, caplen);
                return;
        }
        
        packet_add_header(packet, icmphdr_t, p, len, p_icmp, icmp_hdr);
        caplen -= len; 

        detect_plugins_run(packet, &icmp_plugins_list, "icmp");
        
        SliceAndStoreDataPkt(p + len);
	
	signature_engine_process_packet(signature_engine_get_icmp_root(), packet);
}



static void SliceAndStoreIgmpPkt(u_char *p)
{
        /*
         * Sanity check.
         */
        if ( caplen < sizeof(igmphdr_t) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated IGMP",
                                  "truncated IGMP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(igmphdr_t), caplen);
                return;
        }

        packet_add_header(packet, igmphdr_t, p, sizeof(igmphdr_t), p_igmp, igmp_hdr);
        caplen -= sizeof(igmphdr_t);
        
	detect_plugins_run(packet, &igmp_plugins_list, "igmp");
             
        SliceAndStoreDataPkt(p + sizeof(igmphdr_t));
}



static void SliceAndStoreArpPkt(u_char *p)
{
        /*
         * Sanity check.
         */
        if ( caplen < sizeof(etherarphdr_t) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated ARP",
                                  "truncated ARP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(etherarphdr_t), caplen);
                return;
        }

        packet_add_header(packet, etherarphdr_t, p, sizeof(etherarphdr_t), p_arp, arp_hdr);
        caplen -= sizeof(etherarphdr_t);

	detect_plugins_run(packet, &arp_plugins_list, "arp");
           
        SliceAndStoreDataPkt(p + sizeof(etherarphdr_t));
}



static void SliceAndStoreRarpPkt(u_char *p)
{
        /*
         * Sanity check.
         */
        if ( caplen < sizeof(etherarphdr_t) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated RARP",
                                  "truncated RARP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(etherarphdr_t), caplen);
                return;
        }

        packet_add_header(packet, etherarphdr_t, p, sizeof(etherarphdr_t), p_rarp, arp_hdr);
        caplen -= sizeof(etherarphdr_t);

	detect_plugins_run(packet, &rarp_plugins_list, "rarp");
          
        SliceAndStoreDataPkt(p + sizeof(etherarphdr_t));
}



/*
 * Return -1 on error,
 * 1 if a packet was reassembled.
 * 0 if a fragment was queued.
 */

#define FRAGMENT_QUEUED 0
#define PACKET_REASM 1


static int handle_ip_fragment(iphdr_t *ip, unsigned char *p, unsigned int hlen, unsigned int len) 
{
        int ret;
                
        if ( ((long) (len - hlen)) <= 0 ) {
                prelude_do_report(packet, maybe_faked,
                                  "Invalid fragment data size (too short)",
                                  "Caught a fragment with total len ((len=%u - hlen=%u) = %ld)\n "
                                  "caplen is %u.\n", len, hlen, ((long)len - hlen), caplen);
                return -1;
        }

        assert(((long)(len - hlen)) <= caplen);
             
        /*
         * ip_defrag return > 0 if a packet was reassembled,
         *                 == 0 for fragment queued,
         *                  < 0 on error.
         */
        ret = ip_defrag(ip, p + hlen, len - hlen, &packet->allocated_data);
        if ( ret > 0 ) {
                ip = (iphdr_t *) packet->allocated_data;
                caplen = ret;
                assert((ip->ip_hl * 4) < caplen);

                return PACKET_REASM;
        }

        return ret;
}




static void SliceAndStoreIpPkt(unsigned char *p, proto_enum_t ip_kind)
{
        int ret = -1;
        iphdr_t *ip_hdr;        
        unsigned int off, len, hlen;
        
        if ( caplen < sizeof(iphdr_t) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated IP",
                                  "truncated IP (ip header len=%d > caplen=%u ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(iphdr_t), caplen);
                return;
        }
        

        packet_add_header(packet, iphdr_t, p, sizeof(iphdr_t), ip_kind, ip);
        packet->network_layer_depth = packet->depth;        
        ip_hdr = packet->packet[packet->depth].p.ip;
        
        off = ntohs(ip_hdr->ip_off);
        len = ntohs(ip_hdr->ip_len);
        hlen = ip_hdr->ip_hl * 4;

        if ( hlen < sizeof(iphdr_t) ) {
                prelude_do_report(packet, maybe_faked,
                                  "False IP header len",
                                  "Header len for this IP header is %u when it should be >= %d.\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", len, sizeof(iphdr_t));
                return;
        }
        
        if ( len > caplen ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated IP",
                                  "truncated IP (len=%u > caplen=%u).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", len, caplen);
                return;
        }
        
        if ( hlen > caplen ) {
                prelude_do_report(packet, maybe_faked,
                                  "bad IP header length",
                                  "bad IP header length (hlen=%u > caplen=%u).\n\n ",
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", hlen, caplen);
                return;
        }

        if ( ip_kind == p_ip ) 
                detect_plugins_run(packet, &ip_plugins_list, "ip");
        else
                detect_plugins_run(packet, &ipencap_plugins_list, "ipencap");
                
        off = ntohs(ip_hdr->ip_off);
	if ( off & 0x3fff ) {
                ret = handle_ip_fragment(ip_hdr, p, hlen, len);
                if ( ret == PACKET_REASM ) {
                        packet->depth--;
                        SliceAndStoreIpPkt(packet->allocated_data, ip_kind);
                        return;
                }
                
                else if ( ret < 0 )
                        return;
        }
        
        len = hlen - sizeof(iphdr_t);
        if ( len > 0 )
                SliceAndStoreIpOptions(p + sizeof(iphdr_t), len);    

        caplen -= hlen;

        if ( ret == FRAGMENT_QUEUED ) {
                SliceAndStoreDataPkt(p + hlen);
                return;
        }
        
        
     	switch (ip_hdr->ip_p) {
	case IPPROTO_IP:
                /* Ip encapsulated in Ip */
                SliceAndStoreIpPkt(p + hlen, p_ipencap);
		break;
	case IPPROTO_TCP:
                SliceAndStoreTcpPkt(p + hlen);
		break;
	case IPPROTO_UDP: 
                SliceAndStoreUdpPkt(p + hlen);
		break;
	case IPPROTO_ICMP:
                SliceAndStoreIcmpPkt(p + hlen);
		break;
	case IPPROTO_IGMP:
                SliceAndStoreIgmpPkt(p + hlen);
		break;
	default:
                SliceAndStoreDataPkt(p + hlen);
		break;
	}

        signature_engine_process_packet(signature_engine_get_ip_root(), packet);
}



__inline__ static void switch_ethertype(unsigned short ethertype, unsigned char *p) 
{        
        switch (ethertype) {
	case ETHERTYPE_IP:
                SliceAndStoreIpPkt(p, p_ip);
		break;
	case ETHERTYPE_ARP:
                SliceAndStoreArpPkt(p);
		break;
	case ETHERTYPE_REVARP:
                SliceAndStoreRarpPkt(p);
		break;
	default:
                SliceAndStoreDataPkt(p);
                break;
	}
}



__inline__ static void SliceAndStoreIpv6Pkt(u_char *p) 
{
        
}


void SliceAndStoreEtherPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        
        if ( caplen < sizeof(etherhdr_t) ) {
                prelude_do_report(NULL, maybe_faked,
                                  "truncated ETHER",
                                  "truncated ETHER (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.",
                                  sizeof(etherhdr_t), caplen);
                return;
        }
        
        packet = packet_new(packet_recycler, p);
        if ( ! packet )
                return;
        
        packet_add_header(packet, etherhdr_t, p, sizeof(etherhdr_t), p_ether, ether_hdr);
        caplen -= sizeof(etherhdr_t);
        
	detect_plugins_run(packet, &ether_plugins_list, "ether");
        
	switch_ethertype(ntohs(packet->packet[packet->depth].p.ether_hdr->ether_type),
                         p + sizeof(etherhdr_t));
        
        detect_plugins_run(packet, &all_plugins_list, "all");

        packet_release(packet);
}



#define    NULL_HDRLEN 4

void SliceAndStoreNullPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{          
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;

        if ( caplen < NULL_HDRLEN ) {
                prelude_do_report(NULL, maybe_faked,
                                  "truncated NULL",
                                  "truncated NULL (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", NULL_HDRLEN, caplen);
                return;
        }

        packet = packet_new(packet_recycler, p);
        if ( ! packet )
                return;
        
        caplen -= NULL_HDRLEN;  
        SliceAndStoreIpPkt(p + NULL_HDRLEN, p_ip);

        packet_release(packet);
}




#define PPP_HDRLEN 4

void SliceAndStorePppPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        
        if ( caplen < PPP_HDRLEN ) {
                prelude_do_report(NULL, maybe_faked,
                                  "truncated PPP",
                                  "truncated PPP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", PPP_HDRLEN, caplen);
                return;
        }

        packet = packet_new(packet_recycler, p);
        if ( ! packet )
                return;
        
        caplen -= PPP_HDRLEN;
        SliceAndStoreIpPkt(p + PPP_HDRLEN, p_ip);

        packet_release(packet);
}



#define PPP_BSDI_HDRLEN 24

void SliceAndStorePppBsdosPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{        
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
 
        if ( caplen < PPP_BSDI_HDRLEN ) {
                prelude_do_report(NULL, maybe_faked,
                                  "truncated PPP(bsdos)",
                                  "truncated PPP(bsdos) (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", PPP_BSDI_HDRLEN, caplen);
                return;
        }

        packet = packet_new(packet_recycler, p);
        if ( ! packet )
                return;

        caplen -= PPP_BSDI_HDRLEN;
        SliceAndStoreIpPkt(p + PPP_BSDI_HDRLEN, p_ip);

        packet_release(packet);
}



void SliceAndStoreFddiPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        
        if ( caplen < FDDI_HDRLEN ) {
                prelude_do_report(NULL, maybe_faked,
                                  "truncated FDDI",
                                  "truncated FDDI (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", FDDI_HDRLEN, caplen);
                return;
        }        

        packet = packet_new(packet_recycler, p);
        if ( ! packet )
                return;
        
        caplen -= FDDI_HDRLEN;
        SliceAndStoreIpPkt(p + FDDI_HDRLEN, p_ip);

        packet_release(packet);
}



#define ATM_HDRLEN 8
void SliceAndStoreAtmPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        unsigned short ethertype;
        
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        
        if ( caplen < ATM_HDRLEN ) {
                prelude_do_report(NULL, maybe_faked,
                                  "truncated ATM",
                                  "truncated ATM (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", ATM_HDRLEN, caplen);
                return;
        }        

        packet = packet_new(packet_recycler, p);
        if ( ! packet )
                return;
        
        if (p[0] != 0xaa || p[1] != 0xaa || p[2] != 0x03) {
                p += 20;
                caplen -= 20;
        }

        ethertype = p[6] << 8 | p[7];
        caplen -= ATM_HDRLEN;

        switch_ethertype(ethertype, p + ATM_HDRLEN);
        
        SliceAndStoreIpPkt(p + ATM_HDRLEN, p_ip);

        packet_release(packet);
}



void SliceAndStoreRawPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        const iphdr_t *ip;
        
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;

        ip = (const iphdr_t *) p;
        
        switch(ip->ip_v) {

        case 4:
                packet = packet_new(packet_recycler, p);
                if ( ! packet )
                        return;
                
                SliceAndStoreIpPkt(p, p_ip);
                
                packet_release(packet);
                break;

        case 6:
                log(LOG_ERR, "we don't handle IPV6 yet.\n");
                break;
        }
}



void packet_release(packet_container_t *packet)
{
        if ( --packet->refcount == 0 ) {
                
                free(packet->captured_data);

                if (packet->allocated_data)
                        free(packet->allocated_data);

                free(packet);
                
        }
}



void packet_lock(packet_container_t *packet) 
{
        packet->refcount++;
}
