/*****
*
* Copyright (C) 1998, 1999, 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <dlfcn.h>
#include <errno.h>
#include <limits.h> /* for NAME_MAX */
#include <sys/time.h>

#include "packet.h"

#include "list.h"
#include "plugin-common.h"
#include "plugin-common-prv.h"
#include "plugin-protocol.h"
#include "plugin-detect.h"
#include "common.h"



static LIST_HEAD(protocol_plugins);



/*
 * Subscribe plugins to it's specified protocol,
 * if the plugin need to subscribe to > 1 protocol :
 * we need to copy it.
 */
static int plugin_subscribe(plugin_container_t *pc)
{
        plugin_protocol_t *plugin = (plugin_protocol_t *) pc->plugin;
        
        log(LOG_INFO, "\t\t%s subscribed for \"%s\" protocol handling.\n", plugin->name, plugin->proto);

        INIT_LIST_HEAD(&plugin->detect_plugin_list);
        plugin_register_for_use(pc, &protocol_plugins, plugin->proto);

        return 0;
}



/*
 * Start all plugins of kind 'list'.
 */
int protocol_plugins_run(packet_container_t *packet, unsigned char *data, int len)
{
        int ret = 0;
        struct list_head *tmp;
        plugin_container_t *pc;
         
        list_for_each(tmp, &protocol_plugins) {
                pc = list_entry(tmp, plugin_container_t, ext_list);
                
                plugin_run_with_return_value(pc, &ret, plugin_protocol_t, packet, data, len);
                if ( ret > 0 ) {
                        detect_plugins_run(packet, &((plugin_protocol_t *)pc->plugin)->detect_plugin_list, NULL);
                        break;
                }
        }

        return ret;
}



plugin_protocol_t *protocol_plugins_search(const char *proto) 
{
        struct list_head *tmp;
        plugin_container_t *pc;
        plugin_protocol_t *plugin;
        
        list_for_each(tmp, &protocol_plugins) {         
                pc = list_entry(tmp, plugin_container_t, ext_list);

                plugin = (plugin_protocol_t *) pc->plugin;
                
                if ( strcmp(plugin->proto, proto) == 0 )
                        return (plugin_protocol_t *) pc->plugin;
        }

        return NULL;
}




/*
 * Open the plugin directory (dirname),
 * and try to load all plugins located int it.
 */
int protocol_plugins_init(const char *dirname)
{
        int ret;

        ret = plugin_load_from_dir(dirname, plugin_subscribe);
        if ( ret < 0 ) {
                log(LOG_ERR, "couldn't load plugin subsystem.\n");
                return -1;
        }

        return 0;
}



int protocol_plugin_is_port_ok(port_list_t *plist, uint16_t port) 
{
	int i;

	for ( i = 0; i < plist->entry; i++ ) 
		if ( plist->port_list[i] == port )
			return 0;
	
	return -1;
}
	


int protocol_plugin_add_port_to_list(port_list_t *plist, uint16_t port)
{	
	plist->port_list = realloc(plist->port_list, ++plist->entry * sizeof(uint16_t));
	if ( ! plist->port_list ) {
		log(LOG_ERR, "memory exhausted.\n");
		return -1;
	}
	
	plist->port_list[plist->entry - 1] = port;
	
	return 0;
}



int protocol_plugin_add_string_port_to_list(port_list_t *plist, const char *pstring) 
{
	int ret = -1;
	char *ptr, *s;

	s = ptr = strdup(pstring);
	
        while ( (ptr = strtok(ptr, " ")) ) {

                ret = protocol_plugin_add_port_to_list(plist, atoi(ptr));
		if ( ret < 0 ) 
			break;
		
		ptr = NULL;
        }

        free(s);

        return ret;
}



int protocol_plugin_is_port_list_empty(port_list_t *plist) 
{
        return ( ! plist->entry ) ? 0 : -1;
}



void protocol_plugin_init_port_list(port_list_t *plist) 
{
	memset(plist, 0, sizeof(port_list_t));
}


