/*****
*
* Copyright (C) 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdarg.h>
#include <time.h>

#include "packet.h"
#include "packet-decode.h"
#include "timer.h"
#include "common.h"
#include "alert.h"
#include "alert-prv.h"
#include "plugin-common.h"
#include "rqueue.h"
#include "recycler.h"
#include "plugin-detect.h"
#include "rsend.h"



/*
 * Wait one second after last report before
 * sending it, in order to avoid duplicate on attack
 * where the IP address can't be faked (kind != normal)...
 *
 * This module is bypassed if the report 'kind'
 * is set to normal, because we don't do any check
 * for the attacking source here, and so several source would
 * be merged...
 */


static recycler_t *alert_recycler;
static plugin_generic_t prelude_core_plugin = {
        "Prelude Core", 13,
        "Prelude Core detection engine", 30,
        "The Prelude Team", 17,
        "prelude-devel@lists.sourceforge.net", 36,
};
        

/*
 * Function called back when one of the report queue
 * item expire.
 */
static void rqueue_expire(void *ptr) 
{
        rqueue_t *item = ptr;
        
        timer_destroy(&item->timer);

        log(LOG_INFO, "Debug: Flushing queued report for id=%p, count=%d...\n",
            item, alert_count(item->report));

        /*
         * This is now going to rsend, which will itself
         * be responssible for calling the asynchronous machinery.
         *
         * item->report and packet_release will be called
         * in the async IO thread.
         */
        rsend_emmit(item->report);
        item->report = NULL;
}



/*
 * First time we add this to our report queue.
 */
static void rqueue_first(rqueue_t *item) 
{
        alert_count(item->report) = 1;
        timer_set_expire(&item->timer, 1);
        timer_set_callback(&item->timer, rqueue_expire);
        timer_set_data(&item->timer, item);       
        timer_init(&item->timer);
}



/*
 * Update one of our report queue item.
 */
__inline__ static void rqueue_update(rqueue_t *item) 
{
        alert_time_end(item->report) = time(NULL);
        alert_count(item->report)++;
        timer_reset(&item->timer);
}



/*
 *
 */
static void rqueue_generic(rqueue_t *item) 
{
        /*
         * As the packet will be passed to the async IO subsystem,
         * it should be locked until it is written to Prelude Report.
         */
        packet_lock(item->report->packet);
        
        /*
         * if kind is normal, source IP can't be faked,
         * so we must do one report for one attack...
         * do not queue it.
         */
        if ( item->report->kind == normal ) {
                alert_count(item->report) = 1;
                log(LOG_INFO, "Debug: Flushing report for id=%p, count=%d...\n",
                    item->report, alert_count(item->report));
                rsend_emmit(item->report);
                item->report = NULL;
        } else 
                rqueue_first(item);
}


static int report_new(rqueue_t *item, packet_container_t *packet,
                      rkind_t kind, char *quickmsg, size_t qmsglen) 
{
        alert_t *report;
        
        if ( item->report && kind != normal ) {
                rqueue_update(item);
                return -1;
        } else {
                report = item->report = RecyclerGetChunk(alert_recycler);
                if (! report ) {
                        log(LOG_ERR, "couldn't get a free chunk.\n");
                        return -1;
                }

                RecyclerLockChunk(report);
                
                alert_quickmsg(report) = quickmsg;
                alert_quickmsg_len(report) = qmsglen;
                                
                alert_time_start(report) = time(NULL);
                alert_time_end(report) = 0;
                alert_packet(report) = packet;
                alert_kind(report) = kind;
        }

        return 0;
}



static rkind_t determine_alert_kind(packet_container_t *pc) 
{
        if ( pc->transport_layer_depth < 0 )
                return maybe_faked;

        /*
         * FIXME: use tcp-stream with tcp packet.
         */
        if ( pc->packet[pc->transport_layer_depth].proto == p_tcp )
                return normal;
        else
                return maybe_faked;

#if 0
        ret = tcp_stream_is_known(pc->packet[depth_ip].p.ip,
                                  pc->packet[pc->tcp_depth].p.tcp);

        printf("kind is %s\n", (ret == 0) ? "normal" : "faked");

        return (ret == 0) ? normal : maybe_faked;
#endif
}




/*
 * Queue a report depending on it's kind.
 */
void rqueue_report(rqueue_t *item, plugin_generic_t *plugin,
                   packet_container_t *packet, rkind_t kind,
                   char *qmsg, size_t qmsglen, const char *msg, ...) 
{
        int ret;
        va_list va;

        if ( kind == guess ) {
                kind = determine_alert_kind(packet);

                /*
                 * Special case :
                 * the report queue item currently contain queued report
                 * (not normal kind). We just got a report with kind == normal
                 * using the same report queue item, and so we need to use another
                 * report queue item that'll be flushed right after emission. 
                 */
                if ( kind == normal && item->report ) {
                        static rqueue_t tmp = { NULL };
                        item = &tmp;
                }
        }
        
        ret = report_new(item, packet, kind, qmsg, qmsglen);
        if ( ret < 0 )
                return;
        
        va_start(va, msg);
        ret = vsnprintf(alert_message(item->report),
                        sizeof(alert_message(item->report)), msg, va);
        va_end(va);

        /*
         * return -1 if the output was truncated due to this limit. (Thus until
         * glibc 2.0.6. Since glibc 2.1 these functions follow the
         * C99 standard and return the number of characters (excluding
         * the trailing '\0') which would have been written to
         * the final string if enough space had been available.)
         */
        if ( ret > sizeof(alert_message(item->report)) || ret < 0 )
                alert_message_len(item->report) = sizeof(alert_message(item->report));
        else
                alert_message_len(item->report) = ret + 1;

        alert_plugin(item->report) = plugin ? plugin : &prelude_core_plugin;

        rqueue_generic(item);
}




int rqueue_init(void) 
{
        alert_recycler = RecyclerCreate(sizeof(alert_t), MAX_PKTINUSE);
        if ( ! alert_recycler ) {
                log(LOG_ERR, "couldn't create the alert recycler.\n");
                return -1;
        }

        return 0;
}









