/*****
*
* Copyright (C) 2001 Jeremie Brebec <flagg@ifrance.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
* Written by Jeremie Brebec <flagg@ifrance.com>
*
*****/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>

#include "common.h"
#include "rules.h"
#include "rules_operations.h"
#include "rules_default.h"
#include "rules_parsing.h"
#include "string_matching.h"
#include "rules-type.h"
#include "optparse.h"

#ifdef HAVE_PCRE
#include <pcre.h>
#endif

static int id_ip_src, id_ip_dst;
static int id_port_src, id_port_dst;
static int id_tcp_flags, id_ttl, id_tos, id_id, id_dsize, id_win;
static int id_seq, id_ack, id_ip_proto;
static int id_itype, id_icode, id_icmp_id, id_icmp_seq, id_fragbits;

typedef struct {
	int depth;
	int offset;
	int nocase;
	bm_string_t *s;
} string_t;

static string_t *string_ptr = NULL;

/*
 * Packet matching related function.
 */

static int match_ip_src(packet_container_t *pc, ip_t *test) 
{
        iphdr_t *ip = pc->packet[pc->network_layer_depth].p.ip;
        
        if ( (ip->ip_src.s_addr & test->netmask) == test->ip ) 
                return 0;
        
        return -1;
}



static int match_ip_dst(packet_container_t *pc, ip_t *test) 
{
        iphdr_t *ip = pc->packet[pc->network_layer_depth].p.ip;
        
        if ( (ip->ip_dst.s_addr & test->netmask) == test->ip ) 
                return 0;
        
        return -1;
}



static int match_sameip(packet_container_t *pc, void *data) 
{
        iphdr_t *hdr;

        hdr = pc->packet[pc->network_layer_depth].p.ip;

        return (hdr->ip_src.s_addr == hdr->ip_dst.s_addr) ? 0 : -1;
}




static int match_ip_proto(packet_container_t *pc, integer_t *integer) 
{
        iphdr_t *ip = pc->packet[pc->network_layer_depth].p.ip;

        if ( integer->num == ip->ip_p )
                return 0;

        return -1;
}




static int match_port_src(packet_container_t *pc, segment_t *port) 
{
	int sport, depth;

        depth = pc->transport_layer_depth;
        if ( depth < 0 )
                return -1;

	switch ( pc->packet[depth].proto ) {

        case p_tcp: 
		sport = ntohs(pc->packet[depth].p.tcp->th_sport);
		break;

	case p_udp:
		sport = ntohs(pc->packet[depth].p.udp_hdr->uh_sport);
		break;

	default:
		return -1; 
	}

        if ( sport >= port->lo && sport <= port->hi )
                return 0;

        return -1;
}



static int match_port_dst(packet_container_t *pc, segment_t *port) 
{
	int dport, depth;

        depth = pc->transport_layer_depth;
        if ( depth < 0 )
                return -1;
        
	switch( pc->packet[depth].proto ) {
                
	case p_tcp: 
		dport = ntohs(pc->packet[depth].p.tcp->th_dport);
		break;

	case p_udp:
		dport = ntohs(pc->packet[depth].p.udp_hdr->uh_dport);
		break;

	default:
		return -1; 
	}
        
	if ( dport >= port->lo && dport <= port->hi ) 
                return 0;

        return -1;
}



static int match_tcp_flags(packet_container_t *pc, flags_t *flags) 
{
        int depth = pc->transport_layer_depth;
        
	if ( depth < 0 || pc->packet[depth].proto != p_tcp )
		return -1; 

	if ( (pc->packet[depth].p.tcp->th_flags & flags->mask) == flags->flags )
                return 0;

        return -1;
}



static int match_fragbits(packet_container_t *pc, flags_t *fragbits) 
{
        int depth = pc->network_layer_depth;
        
	if ( (pc->packet[depth].p.ip->ip_p & ~IP_OFFMASK & fragbits->mask) == fragbits->flags )
                return 0;

        return -1;
}



static int match_ttl(packet_container_t *pc, integer_t *ttl) 
{
        int depth = pc->network_layer_depth;
        
	if ( pc->packet[depth].p.ip->ip_ttl == ttl->num )
                return 0;

        return -1;
}



static int match_tos(packet_container_t *pc, integer_t *tos) 
{
        int depth = pc->network_layer_depth;
        
	if ( pc->packet[depth].p.ip->ip_tos == tos->num )
                return 0;

        return -1;
}



static int match_id(packet_container_t *pc, integer_t *id) 
{
        int depth = pc->network_layer_depth;
        
	if ( ntohs(pc->packet[depth].p.ip->ip_id) == id->num )
                return 0;

        return -1;
}



static int match_data_size(packet_container_t *pc, segment_t *dsize) 
{
        int depth = pc->application_layer_depth;
        
	if ( depth < 0 || pc->packet[depth].proto != p_data )
		return -1;

	if ( pc->packet[depth].len >= dsize->lo &&
             pc->packet[depth].len <= dsize->hi )
                return 0;

        return -1;
}



static int match_win(packet_container_t *pc, integer_t *win) 
{
        int depth = pc->transport_layer_depth;

        if ( depth < 0 || pc->packet[depth].proto != p_tcp )
                return -1;

        if ( ntohs(pc->packet[depth].p.tcp->th_win) == win->num )
                return 0;
        
        return -1;
}



static int match_seq(packet_container_t *pc, integer_t *seq) 
{
        int depth = pc->transport_layer_depth;
        
	if ( depth < 0 || pc->packet[depth].proto != p_tcp )
		return -1;
	
	if ( ntohl(pc->packet[depth].p.tcp->th_seq) == seq->num )
                return 0;

        return -1;
}



static int match_ack(packet_container_t *pc, integer_t *ack) 
{
        int depth = pc->transport_layer_depth;
        
	if ( depth < 0 || pc->packet[depth].proto != p_tcp )
		return -1;
	
	if ( ntohl(pc->packet[depth].p.tcp->th_ack) == ack->num )
                return 0;

        return -1;
}



static int match_itype(packet_container_t *pc, integer_t *itype) 
{
        int depth = pc->network_layer_depth + 1;
        
	if ( depth < 0 || pc->packet[depth].proto != p_icmp )
		return -1;

	if ( pc->packet[depth].p.icmp_hdr->icmp_type == itype->num )
                return 0;

        return -1;
}



static int match_icode(packet_container_t *pc, integer_t *icode) 
{
        int depth = pc->network_layer_depth + 1;
        
	if ( depth < 0 || pc->packet[depth].proto != p_icmp )
		return -1;
        
	if ( pc->packet[depth].p.icmp_hdr->icmp_code == icode->num )
                return 0;

        return -1;
}



static int match_icmp_id(packet_container_t *pc, integer_t *test) 
{
        int depth = pc->network_layer_depth + 1;
        
	if ( depth < 0 || pc->packet[depth].proto != p_icmp )
		return -1;

	if ( ntohs(pc->packet[depth].p.icmp_hdr->icmp_id) == test->num )
                return 0;

        return -1;
}



static int match_icmp_seq(packet_container_t *pc, integer_t *test) 
{
        int depth = pc->network_layer_depth + 1;
        
	if ( depth < 0 || pc->packet[depth].proto != p_icmp )
		return -1;

	if ( ntohs(pc->packet[depth].p.icmp_hdr->icmp_seq) == test->num )
                return 0;

        return -1;
}



static int match_ipopts(packet_container_t *pc, void *data) 
{
	flags_t *ipopts = data;
	int depth = pc->network_layer_depth + 1;
        
	if ( pc->packet[depth].proto != p_ipopts )
		return -1;

	return option_is_set(ipopts->flags, 
                             pc->packet[depth].p.opts, 
                             pc->packet[depth].len);
}



#include <sys/time.h>

int signature_engine_match_content(unsigned char *pdata, int dlen, void *data) 
{
        int ret;
	string_t *str;
#if 0
        struct timeval start, end;
#endif           
	str = data;
       			
	if ( str->offset > dlen ) 
		return -1;
		
	pdata += str->offset;
	dlen -= str->offset;

        /*
         * If a "depth" was provided for this content test,
         * set data len to this depth in order to only tests what
         * interest us.
         *
         * if not, depth is 65535, so len will never be greater than.
         */
	if ( dlen > str->depth )
		dlen = str->depth;
        
        if ( dlen < str->s->length ) 
                return -1;

#if 0
        gettimeofday(&start, NULL);
#endif
        
	if ( str->nocase )
		ret = BoyerMoore_CI_StringMatching(pdata, dlen, str->s->pattern, str->s->length,
                                                   str->s->bad_character_shift, str->s->good_suffix_shift);
	else
		ret = BoyerMoore_StringMatching(pdata, dlen, str->s->pattern, str->s->length,
                                                str->s->bad_character_shift, str->s->good_suffix_shift);

#if 0
        gettimeofday(&end, NULL);
        
        log(LOG_INFO, "boyer moorex took %fs to execute.\n", (end.tv_sec + (double) (end.tv_usec * 1e-6)) -
            (start.tv_sec  + (double) (start.tv_usec * 1e-6)));
#endif
        
        return (ret == 0) ? -1 : 0;
}




#ifdef HAVE_PCRE
static int match_pcre(packet_container_t *packet, void *data)
{
        int dlen, ret;
        unsigned char *pdata;

#if 0
        struct timeval start, end;
#endif
        
	if ( packet->packet[packet->depth].proto != p_data )
		return -1;

        dlen = packet->packet[packet->depth].len;
	pdata = packet->packet[packet->depth].p.data;

#if 0
        gettimeofday(&start, NULL);
#endif
        
        ret = pcre_exec((pcre *)data, NULL, pdata, dlen, 0, 0, NULL, 0);

#if 0
        gettimeofday(&end, NULL);

        log(LOG_INFO, "pcre took %fs to execute.\n", (end.tv_sec + (double) (end.tv_usec * 1e-6)) -
            (start.tv_sec  + (double) (start.tv_usec * 1e-6)));
#endif
        
        if ( ret < 0 )
                return -1;
                
	return 0;
}
#endif




static int match_content(packet_container_t *packet, void *data) 
{
        int dlen;
        unsigned char *pdata;

        if ( packet->packet[packet->depth].proto != p_data )
		return -1;

        dlen = packet->packet[packet->depth].len;
        pdata = packet->packet[packet->depth].p.data;
        
        return signature_engine_match_content(pdata, dlen, data);
}




/*
 * Parsing function.
 *
 * TODO :
 * - session:  ?
 * - content-list: 
 * - regex: 
 * - uricontent: (HTTP protocol plugin)
 */

#define DO_PARSE_VALUE(Name, Type)                                                   \
                                                                                     \
static int parse_##Name(char *str, rules_t **rules) {                                \
       Type##_t *type_parsed = parse_##Type(str);                                    \
       if ( type_parsed ) {                                                          \
               *rules = make_new_rules(make_new_rule(id_##Name, type_parsed), NULL); \
               return 0;                                                             \
       } else {                                                                      \
               signature_parser_set_error("Invalid " #Name " value (%s)", str);      \
               *rules = NULL;                                                        \
               return -1;                                                            \
       }                                                                             \
}



static int parse_nothing(parameters_t *parameters, rules_t **rules) 
{
        *rules = NULL;
        return 0;
}


DO_PARSE_VALUE( ttl,      integer );
DO_PARSE_VALUE( tos,      integer );
DO_PARSE_VALUE( id,       integer );
DO_PARSE_VALUE( dsize,    segment );
DO_PARSE_VALUE( win,      integer );
DO_PARSE_VALUE( seq,      integer );
DO_PARSE_VALUE( ack,      integer );
DO_PARSE_VALUE( itype,    integer );
DO_PARSE_VALUE( icode,    integer );
DO_PARSE_VALUE( icmp_id,  integer );
DO_PARSE_VALUE( icmp_seq, integer );
DO_PARSE_VALUE( ip_proto, integer );



static int parse_ip_type(int type_id, char *str, rules_t **rules)
{
	ip_t *ip;

        if ( strcasecmp(str, "any") == 0 ) {
		*rules = NULL;
		return 0;
	}

	ip = parse_ip(str);
	if ( ! ip )
		return -1;

	*rules = make_new_rules(make_new_rule(type_id, ip), NULL);

	return 0;
}



static int parse_ipsrc(char *str, rules_t **rules) 
{
	return parse_ip_type(id_ip_src, str, rules);
}



static int parse_ipdst(char *str, rules_t **rules) 
{
	return parse_ip_type(id_ip_dst, str, rules);
}



static int parse_sameip(parameters_t *parameters, rules_t **rules) 
{
        rule_t *rule;
        
        rule = make_new_rule(0, NULL);
	add_rule_leaf_match(rule, 0, NULL, &match_sameip);
	*rules = make_new_rules(rule, NULL);

        return 0;
}



static int parse_port_type(int type_id, char *str, rules_t **rules) 
{
	char *s, *cstr;
	segment_t *port;

	if ( strncasecmp(str, "any", 3) == 0 ) {
		*rules = NULL;
		return 0;
	}

	port = malloc(sizeof(segment_t));
	if ( ! port ) {
                log(LOG_ERR, "memory exhausted.\n");
		return -1;
	}

	cstr = strdup(str);

	s = strtok(cstr, ":");
	if ( *s != '\0' )
		port->lo = atoi(s);
	else 
		port->lo = 0;

	s = strtok(NULL, "");
	if ( ! s )
		port->hi = port->lo;
	else
		if ( *s == '\0' )
			port->hi = 65535;
	else
		port->hi = atoi(s);

	if ( port->lo < 0 || port->lo > 65535 || port->hi < 0 || port->hi > 65535 ) {
		signature_parser_set_error("Invalid port [%d:%d]", port->lo, port->hi);
		free(port);
		free(cstr);
		return -1;
	}

	*rules = make_new_rules(make_new_rule(type_id, port), NULL);
	free(cstr);

	return 0;
}



static int parse_portsrc(char *str, rules_t **rules) 
{
	return parse_port_type(id_port_src, str, rules);
}



static int parse_portdst(char *str, rules_t **rules) 
{
	return parse_port_type(id_port_dst, str, rules);
}



static int parse_msg(char *str, rules_t **rules) 
{
	data_msg_t *msg;

        msg = malloc(sizeof(data_msg_t));
	if ( ! msg ) {
                log(LOG_ERR, "memory exhausted.\n");
		return -1;
	}
        
	msg->msg = strdup(str);
	msg->len = strlen(str);

	*rules = make_new_rules(make_data_rule(RULE_MESSAGE, msg), NULL);
        
        return 0;

}



static int parse_reference(parameters_t *parameters, rules_t **rules) 
{
        int i = 0;
	char buf[1024], *type;
	data_msg_t *ref;
	static struct {
		const char* name;
		const char* msg;
	} ref_tbl[] = {
		{ "cve",       "(%s)\nSee http://cve.mitre.org/cgi-bin/cvename.cgi?name=%s" },
		{ "bugtraq",   "(%s)\nSee http://www.securityfocus.com/vdb/bottom.html?vid=%s" } ,
		{ "macafee",   "(%s)\nSee http://vil.nai.com/vil/dispVirus.asp?virus_k=%s" },
		{ "arachnids", "(%s)\nSee http://www.whitehats.com/info/IDS%s" },
		{ "url",       "(%s)\nSee %s" },
		{ NULL,        "Reference is unknown\n %s,%s" }
	};

	if ( ! parameters ) {
		signature_parser_set_error("Empty reference field");
		return -1;
	}
	
	if ( parameters->next ) 
		type = parameters->next->str;
	else
		type = "unknown";

	ref = malloc(sizeof(data_msg_t));
	if ( ! ref ) {
                log(LOG_ERR, "memory exhausted.\n");
		return -1;
	}

	while( ref_tbl[i].name != NULL && strcasecmp(ref_tbl[i].name, type) )
		i++;
        
	snprintf(buf, sizeof(buf), ref_tbl[i].msg, type, parameters->str);

	ref->msg = strdup(buf);
	ref->len = strlen(buf);

	*rules = make_new_rules(make_data_rule(RULE_REFERENCE , ref), NULL);

        return 0;
}



/*
 * parse ip options
 *
 * rr - Record route 
 * eol - End of list 
 * nop - No op 
 * ts - Time Stamp 
 * sec - IP security option 
 * lsrr - Loose source routing 
 * ssrr - Strict source routing 
 * satid - Stream identifier 
 */
static int parse_ipopts( char *str, rules_t **rules ) 
{
	rule_t *rule;
        flags_t *flag;
        int i = 0, r = 1;
	static struct {
		const char *name;
		int value;
	} ipopts_tbl[] = {
		{ "rr" ,   IPOPT_RR }, 
		{ "eol" ,  IPOPT_EOL },
		{ "nop",   IPOPT_NOP },
		{ "ts",    IPOPT_TIMESTAMP },
		{ "sec",   IPOPT_SECURITY },
		{ "lsrr",  IPOPT_LSRR },
		{ "lsrre", /*IPOPT_LSRRE*/ 0x84 },
		{ "ssrr",  IPOPT_SSRR },
		{ "satid", IPOPT_SATID },
		{ "ra",    IPOPT_RA },
		{ NULL,    0 }
	};
	
	while( ipopts_tbl[i].name != NULL && strcasecmp(ipopts_tbl[i].name, str) ) {
		i++;
		r = r << 1;
	}

	if ( ! ipopts_tbl[i].name ) {
		signature_parser_set_error("Unknow ipopt '%s'", str);
		return -1;
	}

	flag = malloc(sizeof(flags_t));
	if ( ! flag ) {
                log(LOG_ERR, "memory exhausted.\n");
		return -1;
	}

	flag->mask = r;
	flag->flags = ipopts_tbl[i].value;


	rule = make_new_rule(0, NULL);
	add_rule_leaf_match(rule, RULE_IPOPTS, flag, &match_ipopts);
	*rules = make_new_rules(rule, NULL);

	return 0;
}



static int parse_nocase(parameters_t *parameters, rules_t **rules)
{
	if ( parameters ) {
		signature_parser_set_error("nocase keyword don't take parameters\n");
		return -1;
	}

	if ( ! string_ptr ) {
		signature_parser_set_error("nocase keyword must be used after the content key\n");
		return -1;
	}

	string_ptr->nocase = 1;
	BoyerMoore_MakeCaseInsensitive(string_ptr->s);

        return 0;
}



#ifdef HAVE_PCRE
static int parse_pcre(char *str, rules_t **rules)
{
        pcre *pcre;
        rule_t *rule;
        int erroffset;
	const char *errptr;

	pcre = pcre_compile(str, PCRE_DOLLAR_ENDONLY, &errptr, &erroffset, NULL);
	if ( ! pcre ) {
		signature_parser_set_error("pcre pattern failed to compile.\n");
		return -1;
	}
	
	rule = make_new_rule(0, NULL);
	add_rule_leaf_match(rule, 0, pcre, &match_pcre);
	*rules = make_new_rules(rule, NULL);

	return 0;
}
#endif



static int parse_offset(char *str, rules_t **rules) 
{
        if ( ! string_ptr ) {
		signature_parser_set_error("offset keyword must be used after the content key\n");
		return -1;
        }

	string_ptr->offset = atoi(str);
        return 0;
}



static int parse_depth(char *str, rules_t **rules) 
{
        if ( ! string_ptr ) {
		signature_parser_set_error("depth keyword must be used after the content key\n");
		return -1;
        }

        string_ptr->depth = atoi(str);
	return 0;
}




static int hex2dec(uint8_t c) 
{
        int ret;

        ret = isdigit(c);
        if ( ret )
                return c - '0';

        ret = isalpha(c);
        if ( ret )
                return tolower(c) - 'a' + 10;

        return 0;
}



static int parse_content(const char *str, char *dst, size_t dlen) 
{
        enum { hexa, normal } mode = normal;
        int hexadata = 0, hexalast = 0, cur = 0;
        
        while ( *str != '\0' && cur < (dlen - 1) ) {
                
                switch ( *str ) {
                                
                case '\\':
                        if ( mode == hexa ) {
                                signature_parser_set_error("Invalid \\ in hexadecimal mode.");
                                return -1;
                        }
                        
                        if ( *++str != '\0' )
                                dst[cur++] = *str;
                        break;
                        
                case '|':
                        mode = ( mode == hexa ) ? normal : hexa;
                        break;
                        
                case ' ':
                        if (mode == hexa)
                                break; /* ignore */
                        
                case '0' ... '9':
                case 'a' ... 'f':
                case 'A' ... 'F':
                        if (mode == hexa) {
                                if (hexalast) 
                                        dst[cur++] = (hexadata << 4) + hex2dec(*str);
                                else
                                        hexadata = hex2dec(*str);
                                hexalast = ! hexalast;
                                
                                break;
                        }
                        
                default:
                        if (mode == hexa) {
                                signature_parser_set_error("Invalid character in hexa buffer '%c'.", *str);
                                return -1;
                        } else
                                dst[cur++] = *str;
                        break;
                }
                
		str++;
	}
        
	dst[cur] = '\0';

        return cur;
}




/*
 * Contents functions are public, because there is several place
 * that might use them.
 */
int signature_parser_parse_content(char *str, rules_t **rules)
{
        int ret;
	char dst[2048];
	rule_t *rule;
        
        
	if ( strlen(str) == 0 ) {
		signature_parser_set_error("Missing value for content option.");
		return -1;
	}

        ret = parse_content(str, dst, sizeof(dst));
        if ( ret < 0 )
                return -1;
	
	string_ptr = malloc(sizeof(string_t));
	if ( ! string_ptr ) {
		log(LOG_ERR, "memory exhausted.\n");
		return -1;
	}
	
	string_ptr->s = BoyerMoore_Init(dst, ret);
	if ( ! string_ptr->s )
		return -1;

	string_ptr->offset = 0;
	string_ptr->nocase = 0;
	string_ptr->depth = 65535;

	rule = make_new_rule(0, NULL);
	add_rule_leaf_match(rule, RULE_CONTENT, string_ptr, &match_content);
	*rules = make_new_rules(rule, NULL);

	return 0;
}



static int parse_content_list(char *filename, rules_t **rules) 
{
        int ret;
        FILE *fd;
        char buf[1024];
        rules_t *gbrules = NULL;        

        fd = fopen(filename, "r");
        if ( ! fd ) {
                signature_parser_set_error("couldn't open file %s (%s).\n",
                                           filename, strerror(errno));
                return -1;
        }
        
        while ( fgets(buf, sizeof(buf), fd) ) {
                buf[strlen(buf) - 1] = '\0';

                ret = signature_parser_parse_content(buf, rules);
                if ( ret < 0 )
                        return -1;
                
                printf("buf=\"%s\" gbrules = %p, rules = %p\n", buf, gbrules, *rules);
                gbrules = rules_or(gbrules, *rules);
        }

        fclose(fd);

        *rules = gbrules;

        return 0;
}




/*
 * parse fragbits
 *
 * R - Reserved bit
 * D - DF bits
 * M - MF bits
 */
static int parse_fragbits( char *str, rules_t **rules ) 
{
        int flags = 0;
	rules_t *new_rules = NULL;
        enum { mode_all, mode_any, mode_exact } mode = mode_exact;
        
	if ( strlen(str) == 0 ) 
		return -1;
        
	while( *str != '\0' ) {
		switch (*str) {
                        
		case 'r':
		case 'R': 
			flags |= IP_RF;
			break;
                        
		case 'd':
		case 'D':
			flags |= IP_DF;
			break;
                        
		case 'm':
		case 'M':
			flags |= IP_MF;
			break;
                        
		case '0':
			flags = 0; /* snort compatibility ? */
			break;
                        
		case '+':
			mode = mode_all;
			break;
                        
		case '*':
			mode = mode_any;
			break;
                        
		default: 
			signature_parser_set_error("Unknow fragbits flag %c", *str);
			return -1;
		}
		str++;
	}

	/* make rule */
	if ( mode == mode_any ) {
		/* 
		 * decompose flags on x test "all"
		 *  AS+ = A* and S*
		 */
		int i;
                
		new_rules = make_new_rules(NULL, NULL);

		for (i = 1; i < 256; i = i << 1)
			if ( flags & i ) {
				rule_t *rule;
				flags_t *flag;

                                flag = malloc( sizeof(flags_t) );
				if ( ! flag ) {
                                        log(LOG_ERR, "memory exhausted.\n");
                                        return -1;
				}
                                
				flag->mask = i;
				flag->flags = i;
				
				rule = make_new_rule(id_fragbits, flag);
				new_rules = rule_and(new_rules, rule);
			}
	} else {
		flags_t *flag;

                flag = malloc(sizeof(flags_t));
		if ( ! flag ) {
			log(LOG_ERR, "memory exhausted.\n");
                        return -1;
		}
                
		flag->mask = (mode == mode_all) ? flags : ~0;
		flag->flags = flags;
		
		new_rules = make_new_rules(make_new_rule(id_fragbits, flag), NULL);
	}

	*rules = new_rules;

        return 0;
}




/*
 * parse tcp flags:
 *
 * F - FIN (LSB in TCP Flags byte)
 * S - SYN
 * R - RST
 * P - PSH
 * A - ACK
 * U - URG
 * 2 - Reserved bit 2
 * 1 - Reserved bit 1 (MSB in TCP Flags byte)
 
 * There are also logical operators that can be used to specify matching criteria for the
 * indicated flags:
 * + - ALL flag, match on all specified flags plus any others
 * * - ANY flag, match on any of the specified flags
 *
 */

#ifndef TH_RES1
#define TH_RES2         0x40
#define TH_RES1         0x80                                                                            
#endif

static int parse_tcp_flags(char *str, rules_t **rules) 
{
	int flags = 0;
	rules_t *new_rules = NULL;
        enum { mode_all, mode_any, mode_exact } mode = mode_exact;

	if ( strlen(str) == 0 ) 
		return -1;

	/* parse flags string */
	while( *str != '\0' ) {
		switch( *str ) {
		case 'f':
		case 'F': 
			flags |= TH_FIN;
			break;
                        
		case 's':
		case 'S':
			flags |= TH_SYN;
			break;
                        
		case 'r':
		case 'R':
			flags |= TH_RST;
			break;
                        
		case 'p':
		case 'P':
			flags |= TH_PUSH;
			break;
                        
		case 'a':
		case 'A':
			flags |= TH_ACK;
			break;
                        
		case 'u':
		case 'U':
			flags |= TH_URG;
			break;
                        
		case '1':
			flags |= TH_RES1;
			break;
                        
		case '2':
			flags |= TH_RES2;
			break;
                        
		case '0':
			flags = 0; /* snort compatibility ? */
			break;
                        
		case '+':
			mode = mode_all;
			break;
                        
		case '*':
			mode = mode_any;
			break;
                        
		default: 
			signature_parser_set_error("Unknow tcp flag %c", *str);
			return -1;
		}
		str++;
	}

	/* make rule */
	if ( mode == mode_any ) {
		/* 
		 * decompose flags on x test "all"
		 *  AS+ = A* and S*
		 */
		int i;
                
		new_rules = make_new_rules(NULL, NULL);

		for ( i = 1; i < 256; i = i << 1 )
			if (flags & i) {
				rule_t *rule;
				flags_t *flag;

                                flag = malloc(sizeof(flags_t));
				if ( ! flag ) {
					log(LOG_ERR, "memory exhausted.\n");
					return -1;
				}
                                
				flag->mask = i;
				flag->flags = i;
                                
				rule = make_new_rule(id_tcp_flags, flag);
				new_rules = rule_and(new_rules, rule);
			}
	} else {
		flags_t *flag;

                flag = malloc( sizeof(flags_t) );
		if ( ! flag ) {
                        log(LOG_ERR, "memory exhausted.\n");
                        return -1;
		}
                
		flag->mask = (mode == mode_all) ? flags : ~0;
		flag->flags = flags;
		new_rules = make_new_rules(make_new_rule(id_tcp_flags, flag), NULL);

	}

	*rules = new_rules;
	return 0;
}




/*
 * Register defaults rules type
 */



static int set_nocase(rules_t *rules) 
{
	string_ptr = NULL;
	
	return 0;
}



/**********************************************************
 * Register defaults rules type
 **********************************************************/

static rules_node_t *ip_root, *tcp_root, *udp_root, *icmp_root;

/* TODO, priority statistics */ 

/* BASE */
#define PRIORITY_IP_DST         0
#define PRIORITY_PORT_DST       2
#define PRIORITY_IP_SRC         4
#define PRIORITY_PORT_SRC       6

#define PRIORITY_IPPROTO        7

/* IP */
#define PRIORITY_ID            11
#define PRIORITY_TOS           12
#define PRIORITY_TTL           13
#define PRIORITY_FRAGBITS      8
#define PRIORITY_IPOPTIONS     9
#define PRIORITY_DSIZE         25

/* ICMP */
#define PRIORITY_ITYPE         20
#define PRIORITY_ICODE         21
#define PRIORITY_ICMP_ID       27
#define PRIORITY_ICMP_SEQ      28

/* TCP */
#define PRIORITY_TCP_FLAGS     10
#define PRIORITY_SEQ           27
#define PRIORITY_ACK           28
#define PRIORITY_WIN           25


void signature_engine_init(void)
{
	/*
         * register default type
         */
	id_ip_src = REGISTER_TYPE(PRIORITY_IP_SRC, ip, match_ip_src);
	id_ip_dst = REGISTER_TYPE(PRIORITY_IP_DST, ip, match_ip_dst);
	id_port_src = REGISTER_TYPE(PRIORITY_PORT_SRC, segment, match_port_src);
	id_port_dst = REGISTER_TYPE(PRIORITY_PORT_DST, segment, match_port_dst);
	id_tcp_flags = REGISTER_TYPE(PRIORITY_TCP_FLAGS, flags, match_tcp_flags);
	id_ttl = REGISTER_TYPE(PRIORITY_TTL, integer, match_ttl);
	id_tos = REGISTER_TYPE(PRIORITY_TOS, integer, match_tos);
	id_id = REGISTER_TYPE(PRIORITY_ID, integer, match_id);
	id_dsize = REGISTER_TYPE(PRIORITY_DSIZE, segment, match_data_size); 
	id_win = REGISTER_TYPE(PRIORITY_WIN, integer, match_win);
	id_seq = REGISTER_TYPE(PRIORITY_SEQ, integer, match_seq);
	id_ack = REGISTER_TYPE(PRIORITY_ACK, integer, match_ack);
	id_itype = REGISTER_TYPE(PRIORITY_ITYPE, integer, match_itype);
	id_icode = REGISTER_TYPE(PRIORITY_ICODE, integer, match_icode);
	id_icmp_id = REGISTER_TYPE(PRIORITY_ICMP_ID, integer, match_icmp_id);
	id_icmp_seq = REGISTER_TYPE(PRIORITY_ICMP_SEQ, integer, match_icmp_seq);
	id_fragbits = REGISTER_TYPE(PRIORITY_FRAGBITS, flags, match_fragbits);
        id_ip_proto = REGISTER_TYPE(PRIORITY_IPPROTO, integer, match_ip_proto);
        
	/*
         * register default key
         */
	signature_parser_add_one_arg_key("ipsrc", &parse_ipsrc);
	signature_parser_add_one_arg_key("ipdst", &parse_ipdst);
        signature_parser_add_one_arg_key("ip_proto", &parse_ip_proto);
        signature_parser_add_one_arg_key("portsrc", &parse_portsrc);
        signature_parser_add_one_arg_key("portdst", &parse_portdst);
        signature_parser_add_one_arg_key("msg", &parse_msg);
        signature_parser_add_one_arg_key("flags", &parse_tcp_flags);
        signature_parser_add_one_arg_key("ttl", &parse_ttl);
        signature_parser_add_one_arg_key("tos", &parse_tos);
        signature_parser_add_one_arg_key("id", &parse_id);
        signature_parser_add_one_arg_key("dsize", &parse_dsize);
	signature_parser_add_one_arg_key("window", &parse_win);
        signature_parser_add_one_arg_key("seq", &parse_seq);
	signature_parser_add_one_arg_key("ack", &parse_ack);
        signature_parser_add_one_arg_key("itype", &parse_itype);
        signature_parser_add_one_arg_key("icode", &parse_icode);
        signature_parser_add_one_arg_key("icmp_id", &parse_icmp_id);
        signature_parser_add_one_arg_key("icmp_seq", &parse_icmp_seq);
        signature_parser_add_one_arg_key("content", &signature_parser_parse_content);
        signature_parser_add_one_arg_key("content-list", &parse_content_list);
        signature_parser_add_one_arg_key("offset", &parse_offset);
        signature_parser_add_one_arg_key("depth", &parse_depth);
        signature_parser_add_one_arg_key("ipopts", &parse_ipopts);
        signature_parser_add_one_arg_key("fragbits", &parse_fragbits);
#ifdef HAVE_PCRE
	signature_parser_add_one_arg_key("pcre", &parse_pcre);
#endif

        /*
         * Register key that don't take any argument.
         */
        signature_parser_add_multiple_args_key("reference", &parse_reference);
	signature_parser_add_multiple_args_key("nocase", &parse_nocase);
        signature_parser_add_multiple_args_key("sameip", &parse_sameip);

        /*
         * Key to be ignored.
         */
        signature_parser_add_multiple_args_key("classtype", &parse_nothing);
        signature_parser_add_multiple_args_key("sid", &parse_nothing);
        signature_parser_add_multiple_args_key("rev", &parse_nothing);
        signature_parser_add_multiple_args_key("react", &parse_nothing);
        signature_parser_add_multiple_args_key("resp", &parse_nothing);
        signature_parser_add_multiple_args_key("logto", &parse_nothing);

	/* add default post processing function */
        signature_parser_add_post_processing(validate_rules);
	signature_parser_add_post_processing(set_nocase);

	/* make tree */
        ip_root = make_rules_root();
	tcp_root = make_rules_root();
	udp_root = make_rules_root();
	icmp_root = make_rules_root();
}



/*
 * get build-in trees
 */
rules_node_t *signature_engine_get_ip_root(void)
{
        return ip_root;
}


rules_node_t *signature_engine_get_tcp_root(void)
{
        return tcp_root;
}


rules_node_t *signature_engine_get_udp_root(void)
{
        return udp_root;
}


rules_node_t *signature_engine_get_icmp_root(void)
{
        return icmp_root;
}

